/***************************************************************\
*	athresh.c		  				*
*   real-time monitor of audio port, report above threshold	*
*   dpwe 1994mar30						*
\***************************************************************/

/* std libs */
#include <stdlib.h>
#include <string.h>
#include <math.h>	/* for atof ?? */
#include <time.h>
/* personal libs */
#include <dpwelib.h>
#include <genutils.h>
#include <audIO.h>

/* constants */
#define DFLT_THRSH	200	/* default sample peak (out of 32768) */
#define DFLT_TIME	(0.1)	/* default time window */

#define DFLT_SRATE	(16000.0)	/* run sampling at this rate */
#define CHANS		1		/* monitor this many chans */

#define BUFLEN 16384

/* static function prototype */
static void usage(void);

/* static data */
       char  *programName;
static int   verbose = 0;		/* -v flag */
static int   contin  = 0;
static int   peakFilt = 0;	/* try to only react once per peak */
static int   thresh = DFLT_THRSH;
static float srate  = DFLT_SRATE;
static float timew  = DFLT_TIME;

/* functions */

char *timestring(void)
{	/* just return a printable version of the current string */
static char timestr[64];
    time_t tm;

    time(&tm);
    strcpy(timestr, ctime(&tm));
    timestr[strlen(timestr)-1] = '\0';
    return timestr;	/* yuk */
}

void shVLims(short *buf, int sinc, long len, int *pmin, int *pmax, int *pavg)
{	/* find the max and min of a buffer of shorts */
    int min, max, d;
    long sum = 0;
    long l = len;

    min = max = *buf;
    while(l--)  {
	d = *buf;
	if(min > d)		min = d;	
	else if(max < d)	max = d;
	buf += sinc;
	sum += d;
    }
    *pmin = min;
    *pmax = max;
    *pavg = (int)sum/len;
}

void DoProcess(void)
{  /* actually do something */
    long chunksize = srate * timew;
    AUSTRUCT *au;
    short *sbuf;
    int min, max, avg;
    int abvThresh = 0;

    if((au=AUOpen(SFMT_SHORT, srate, CHANS, 2*chunksize, AIO_INPUT))==NULL) {
	fprintf(stderr, "failed to open audio input\n");
	exit(1);
    }
    sbuf = TMMALLOC(short, chunksize*CHANS, "athresh:sbuf");
    while(1) {
	AURead(au, sbuf, chunksize);
	shVLims(sbuf, CHANS, chunksize, &min, &max, &avg);
	max -= avg;	min -= avg;
	if(contin)	/* report average-offset levels each cycle */
	    printf("min = %d, max = %d\n", min, max);
	if(max > thresh || min < -thresh)  {
	    if(!peakFilt || !abvThresh)  {
		abvThresh = 1;
		if(verbose)  {
		    printf("%s ",timestring());
		}
		printf("%d\n", MAX(max, -min));
		fflush(stdout);
	    }
	} else {
	    abvThresh = 0;
	}
    }
}

main(int argc, char **argv)
{
    int 	i = 0;
    int 	err = 0;
    char	*inPath = NULL, *outPath = NULL;

    programName = argv[0];
/*
    strcpy(defext, DFLT_EXTN);
    strcat(defext, sndfExtn);
 */
    i = 0;
    while(++i<argc && err == 0)	 {
	char c, *token, *arg, *nextArg;
	int  argUsed;

	token = argv[i];
	if(*token++ == '-')  {
	    if(i+1 < argc)	nextArg = argv[i+1];
	    else		nextArg = "";
	    argUsed = 0;
	    while(c = *token++)  {
		if(NumericQ(token))	arg = token;
		else			arg = nextArg;
		switch(c)  {
		case 'h':	thresh   = atoi(arg); argUsed = 1; break;
		case 'w':	timew    = atof(arg); argUsed = 1; break;
		case 'r':	srate    = atof(arg); argUsed = 1; break;
		case 'p': 	peakFilt = 1; break;
		case 'c': 	contin   = 1; break;
		case 'v': 	verbose  = 1; break;
		default:	fprintf(stderr,"%s: unrec option %c\n",
					programName, c);
		                err = 1; break;
		}
		if(argUsed)  {
		    if(arg == token)	token = "";   /* no more from token */
		    else		++i;	      /* skip arg we used */
		    arg = ""; argUsed = 0;
		}
	    }
	} else {
/*
	    if(inPath == NULL)		inPath = argv[i];
	    else if(outPath == NULL)	outPath = argv[i];
	    else */
	    { fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
		   err = 1;  }
	}
    }
    /* other tests on args */
/*
    if(data < 0)  { 
	fprintf(stderr,"%s: bad data %d: must be >= 0\n", programName, data);
	err = 1; 
    }
 */
    if(err /* || inPath == NULL */)
	usage();		/* never returns */

/*    if(outPath == NULL)  {
	outPath=TMMALLOC(char, strlen(inPath)+strlen(defext), "outPath");
	strcpy(outPath, inPath);
	SubstExtn(inPath, defext, outPath);
    }
 */
    if(verbose)  {
	fprintf(stderr,"%s: threshold = %d\n",programName,thresh);
	fprintf(stderr,"%s: time win  = %.3f\n",programName,timew);
	fprintf(stderr,"%s: sample rt = %.0f\n",programName,srate);
	fprintf(stderr,"%s: peak filt = %s\n",programName,peakFilt?"yes":"no");
	fprintf(stderr,"%s: continuous= %s\n",programName,contin?"yes":"no");
	fprintf(stderr,"%s: verbose   = %s\n",programName,verbose?"yes":"no");
    }
    DoProcess();
}

static void usage(void)		/* print syntax & exit */
{
    fprintf(stderr,
	    "usage: %s [-h thr][-w time][-r rate][-p][-c][-v]\n",programName);
    fprintf(stderr,"where (defaults in brackets)\n");
    fprintf(stderr," -h thr   level trigger threshold   (%d)\n",DFLT_THRSH);
    fprintf(stderr," -w time  time window               (%.3f s)\n",DFLT_TIME);
    fprintf(stderr," -r rate  input hardware samp. rate (%.0f Hz)\n",DFLT_SRATE);
    fprintf(stderr," -p       attempt to react only once per 'peak'\n");
    fprintf(stderr," -c       print min and max continuously\n");
    fprintf(stderr," -v       verbose mode\n");
/*    fprintf(stderr," infile   input file\n");
    fprintf(stderr," outfile  output file          (inStem+'%s')\n",defext);
 */
    exit(1);
}

